from abc import ABC, abstractmethod


class PrimitiveBase(ABC):
    """
    A base class for all TA1 primitives.
    """


class DataCleaningPrimitiveBase(PrimitiveBase):
    
    def __init__(self):
        """ Initializes the featurization primitive.
        All primitives should specify all the hyper-parameters that can be set 
        at the class level in their ``__init__`` as explicit keyword arguments 
        (no ``*args`` or ``**kwargs``). Available hyper-parameters should be 
        specified in primitive’s D3M annotation.
        """
        self._out_type = None

    @property
    def out_type(self):
        """ This definition of this object allows the primitive to define and
        return an arbitrary structure, which is essential for primitives that
        are augmenting, transforming, and annotating data.

        Returns
        -------
        {
            'type': 'dict',
            'definition': {
                'my_property_1': {
                    'type': 'dict',
                    'my_sub_property_1': {
                        'type': 'label[]'
                    }
                }
            }
        }

        Defines a data structure such as: 

        {
            'my_property_1': {
                'my_sub_property_1': ['a','b','c']
            }
        }

        When the primitive returns a matrix, the keys refer to column positions.
        For example:

        {
            'type': 'matrix',
            'definition': {
                0: {
                    'type': 'label[]'
                },
                1: {
                    'type': 'dict[]',
                    'my_sub_property_1': {
                        'type': 'integer'
                    }
                }
            }
        }

        This would define a data structure such as:

        [
            [
                'a',
                'b',
                'c'
            ],
            [
                {'my_sub_property_1': 1},
                {'my_sub_property_1': 2},
                {'my_sub_property_1': 3}
            ]
        ]

        All sub property definition should conform to the allowed `fit` model
        intype.
        """
        return self._out_type


    @abstractmethod
    def fit(self, intype, data):
        """ Takes input data of the specified type and performs any computation
        required for transformation.

        Parameters
        ----------
        intype : array-like
            A matrix specifying the format of the input data.

            Possible intype formats are:
            "float" : float

            "integer" : integer

            "text" : str
                Document (longer text)

            "label" : string
                1-2 word category label

            "dateTime" : datetime or str
                Python datetime object, or dateTime-formatted string

            "location" : str
                Represents a real-world location

            "coordinatePair" : str
                Latitude, longitude pair, e.g. "10.27,-30.45"

            "matrix" : array-like
                Numpy.ndarray or list.

            "dataset": array-like
                Tabular data structure, where the first row is a list of
                column headers

            Each primitive should specify applicable input data formats in its
            documentation and primitive annotation. The intype property can be
            arbitrarily complex, including nested structures. 
            For example: 

            ["matrix, [""label", "dateTime", ["float","float"]]]

            Would correspond to the structure:

            [
                ["green", "2017-01-01", [1.25,3.5]],
                ["blue", "2017-02-01", [3.25,5.5]]
            ]

        data : array-like
            The input in the specified format. 
        """


    @abstractmethod
    def transform(self, data=None):
        """ Performs the transformation and returns the transformed data.

        Parameters
        ----------
        data : array-like
            optional argument, used in cases where the original input data was
            used to train model, and transformation will now be performed on a
            new set of data

        Returns
        -------
        transformed_data : array-like
        """


    def staged_fit(self, intype, data):
        """ An iterative version of ``fit`` which yields at each internal
        iteration, for example, after every batch of training. This allows
        incremental monitoring and evaluation of the training process. When a
        fit generator yields, it should be possible to call methods on the
        instance and operate on current state of the model (at the time of
        yielding), including pickling the instance.
        """
        yield self.fit(self, intype, data)


    def staged_transform(self, data=None):
        """ An iterative version of ``transform`` which yields at each internal
        iteration. At every iteration, returned value should be of the same shape
        as the return value for ``transform``, but some elements might be
        missing.
        """
        yield self.transform(data=data)



import numpy as np
from sklearn.decomposition import PCA
from ..base import DataCleaningPrimitiveBase


class PCAFeatures(DataCleaningPrimitiveBase):

    def fit(self, intype, data):
        """ Do PCA and return the ranked features.                                  
                                                                                
        The matrix M will correspond to the absolute value of the components of 
        the decomposiiton thus giving a matrix where each column corresponds to 
        a principal component and rows are the components that rescale each 
        feature.  
                                                                                
        For example, pca.components_.T could look like this:                        
                                                                                
        array([[ 0.52237162,  0.37231836, -0.72101681, -0.26199559],                
            [-0.26335492,  0.92555649,  0.24203288,  0.12413481],                 
            [ 0.58125401,  0.02109478,  0.14089226,  0.80115427],                 
            [ 0.56561105,  0.06541577,  0.6338014 , -0.52354627]])                
                                                                                
        So the most important feature with respect to the first principal 
        component would be the 3rd feature as this has an absolute value of 
        0.58125401 which is greater than all the entries in that column.                             
                                                                                
                                                                                
        "importance_on1stpc" corresponds to the indices of most important       
        features for the first principal. Component are in ascending order      
        (most important feature 0, least important feature n_features-1).       
                                                                                
        "importance_onallpcs" corresponds to the indices of the one most        
        important feature for each principal components.                        
                                                                                
        "explained_variance_ratio" Percentage of variance explained by each of  
        the selected components.                                                
        
        Params 
        ------- 
        intype : str
            Expect ``matrix``.
        data : array-like, [n_samples, n_features]
            Training data.
        """  
        if isinstance(intype, (list, tuple)):
            assert(intype[0]=="matrix")
        else:
            raise ValueError("Fit expected a 'matrix' an a intype.")

        pca = PCA()
        pca.fit_transform(data)

        self.components_ = pca.components_
        self.explained_variance_ratio_ = pca.explained_variance_ratio_

        M = np.absolute(pca.components_.T)
        # Rank features based on contribtuions to 1st PC
        self.importance_on1stpc = np.argsort(M[:,0], axis=0)[::-1]
        # Rank features based on contributions to PCs 
        self.importance_onallpcs = np.argmax(M, axis=0)

        return self

    def transform(self, data=None):
        return {
            "importance_on1stpc": self.importance_on1stpc,
            "importance_onallpcs": self.importance_onallpcs
        }



import numpy as np
from sklearn.decomposition import PCA, FactorAnalysis
from sklearn.model_selection import cross_val_score
from ..base import DataCleaningPrimitiveBase


class HeteroscedasticityTest(DataCleaningPrimitiveBase):
    __test__ = False

    def __init__(self, max_iter=1000000, tol=1e-8):
        self.max_iter = max_iter
        self.tol      = tol


    def fit(self, intype, data):
        """ Test heteroscedaticity of your data.

        The consequence is that the likelihood of new data can be used for 
        model selection and covariance estimation.

        Params
        ------  
        data : array-like
            Training data.

        Returns
        -------
        "pca" returns a two tuple of number of componenets and likelihood for 
        best principal componenets estimator.                                   
        "fa" returns a two tuple of number of componenets and likelihood for 
        best factor analysis estimator.
        """
        if isinstance(intype, (list, tuple)):
            assert(intype[0]=="matrix")                   
        else:                                                                   
            raise ValueError("Fit expected a 'matrix' as intype.")

        self.pca_scores, self.fa_scores = self.compute_scores(data, self.max_iter, self.tol)

        self.pca = max(self.pca_scores, key=lambda s: s[0])
        self.fa  = max(self.fa_scores, key=lambda s: s[0])

        return self


    def compute_scores(self, X, max_iter=1000000, tol=1e-6):                              
        """ Compare PCA against FactorAnalysis.                                     
                                                                                
        Code taken from 'http://scikit-learn.org/stable/auto_examples/'             
        'decomposition/plot_pca_vs_fa_model_selection.html'.                        
                                                                                
        Params                                                                      
        ------                                                                      
        X : array-like                                                              
            Training data.                                                          
                                                                                
        Returns                                                                     
        -------                                                                     
        pca_scores, fa_scores : tuple of arrays                                     
        """                                                                         
        pca = PCA(svd_solver='full')                                                
        fa = FactorAnalysis(max_iter=max_iter, tol=tol)                             
                                                                                
        pca_scores, fa_scores = [], []                                              
        n_featues = X.shape[1]                                                      
        if n_featues > 20:                                                          
            n_components = np.arange(0, n_featues+1, 5)                             
        else:                                                                       
            n_components = np.arange(0, n_featues+1)                                
                                                                                
        for n in n_components:                                                      
            pca.n_components = n                                                    
            fa.n_components = n                                                     
            pca_scores.append( (np.mean(cross_val_score(pca, X)), n) )              
            fa_scores.append( (np.mean(cross_val_score(fa, X)), n) )                
                                                                                
        return pca_scores, fa_scores


    def transform(self, data=None):                                            
        return {"pca": self.pca, "fa": self.fa}



import numpy as np
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.pipeline import Pipeline                                                                
from sklearn.model_selection import GridSearchCV
from ..base import DataCleaningPrimitiveBase


class RFFeatures(DataCleaningPrimitiveBase):

    def __init__(self, problem_type, scoring, cv=3, verbose=0, n_jobs=1):
        """
        Params
        ------
        problem_type : str
            ``classification`` or ``regression``.

        cv : int, cross-validation generator or an iterable, optional
            Determines the cross-validation splitting strategy. Possible inputs
            for cv are:
                None, to use the default 3-fold cross validation,
                integer, to specify the number of folds in a (Stratified)KFold,
                An object to be used as a cross-validation generator.
                An iterable yielding train, test splits.

            For integer/None inputs, if the estimator is a classifier and y is
            either binary or multiclass, StratifiedKFold is used. In all other
            cases, KFold is used.

            Refer User Guide for the various cross-validation strategies that
            can be used here.

        n_jobs : int, default=1
            Number of jobs to run in parallel.

        scoring : string, callable, list/tuple, dict or None, default: None
            A single string (see The scoring parameter: defining model
            evaluation rules) or a callable (see Defining your scoring strategy
            from metric functions) to evaluate the predictions on the test set.

            For evaluating multiple metrics, either give a list of (unique)
            strings or a dict with names as keys and callables as values.

            NOTE that when using custom scorers, each scorer should return a
            single value. Metric functions returning a list/array of values can
            be wrapped into multiple scorers that return one value each.
            
            See Specifying multiple metrics for evaluation for an example.
            
            If None, the estimator’s default scorer (if available) is used.

        verbose : integer
            Controls the verbosity: the higher, the more messages.
        """
        self.problem_type = problem_type
        self.scoring = scoring
        self.cv      = cv
        self.n_jobs  = n_jobs
        self.verbose = verbose


    def fit(self, intype, data):
        """ Rank features using Random Forest classifier.                           
                                                                                
	Use GridSearchCV to optimize the scoring method for a random forest         
    	classifier and return the importance of the features.    

        "indices" indicate the feature importances of features order in
        ascending order (most importance feature to least important features).

        "feature_importances" The feature importances (the higher, the more
        important the feature).
                                                                                
    	Params                                                                      
    	------                                                                      
        intype : list or tuple
            Expects ["matrix", "matrix"]
        data : tuple of arrays 
            ([n_samples, n_features], [n_samples, n_features]) corresponding to
            training data and labels.
        """
        # Check data format
        if isinstance(intype, (list, tuple)):
            assert(intype[0]=="matrix")
            assert(intype[1]=="matrix")
        else:
            raise ValueError("Fit expected Two 'matrix's as intype.")
        
        # Unpack data
        X, y = data

        # Chose classfier or regressor
        if self.problem_type=="classification":
            rf = Pipeline([('clf', RandomForestClassifier(random_state=1))])    
        elif self.problem_type=="regression":
            rf = Pipeline([('clf', RandomForestRegressor(random_state=1))])
        else:
            raise ValueError(
                "problem_type must be 'classification' or 'regression'."
            )

        # Gridsearch for hyperparam optimization
        param_grid = [{'clf__n_estimators': [10, 100, 1000, 10000]},]                                          
        gs_rf = GridSearchCV(rf,                                                    
                             param_grid, 
                             cv      = self.cv,
                             scoring = self.scoring,                                     
                             verbose = self.verbose,                                           
                             n_jobs  = self.n_jobs)                                           
        gs_rf.fit(X, y)                                                             
                                                                                
        # Rank from most to least important features (0, d-1)                       
        self.feature_importances = \
                gs_rf.best_estimator_.named_steps["clf"].feature_importances_
        self.indices = np.argsort(self.feature_importances)[::-1]                         

        return self


    def transform(self, data=None):
        return self.indices



from setuptools import setup, find_packages


setup(
    name="punk",

    version="1.0.2",

    description="Primitives for Uncovering New Knowledge.",
    long_description="Machine Learning pipeline elements.",

    url="https://github.com/NewKnowledge/punk",

    author="New Knowledge",
    author_email="support@newknowledge.io",

    license="MIT",

    classifiers=[
        "Development Status :: 4 - Beta",

        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.6",

        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],

    keywords="TA1 primitive, feature selection, novelty detection",


    packages=find_packages(exclude=['tests']),

    install_requires=["numpy", "scikit-learn", "scipy"],


)
